# utils
from sys import argv as options
from backscattering_analyzer.settings import Settings
from backscattering_analyzer.signal import Signal
from backscattering_analyzer import interpolate, interpolate_abciss
from numpy import loadtxt
from scipy.io.matlab import loadmat

# maths
from numpy import zeros, pi, sqrt


class Analyzer:
    """
    Utility class to study backscattering light in VIRGO
    """

    def __init__(self, arguments: list = []) -> None:
        self.settings = Settings(options[1:] + arguments)
        self.load()
        self.process_movement()

    def load(self):
        """
        Load all data
        """
        self.settings.log("Load all data")
        try:
            self.load_bench()
            self.load_mirror()
            self.load_data()
            self.load_reference()
            self.load_coupling()
        except Exception:
            raise Exception("An error occured during data loading")

    def load_bench(self):
        """
        Load bench movement
        """
        file = self.settings.bench_file()
        self.settings.log("loading bench movement")
        try:
            data = loadtxt(file).T
            self.bench_movement = Signal(
                data[0], data[1], self.settings
            )
        except OSError:
            raise Exception("{file} does not exist".format(file=file))

    def load_mirror(self):
        """
        Load mirror movement
        """
        file = self.settings.mirror_file()
        self.settings.log("loading mirror movement")
        try:
            data = loadtxt(file).T
            self.mirror_movement = Signal(
                data[0], data[1], self.settings
            )
        except OSError:
            raise Exception("{file} does not exist".format(file=file))

    def load_data(self):
        """
        Load excited h(t)
        """
        file = self.settings.data_file()
        self.settings.log("loading excited h(t)")
        try:
            data = loadtxt(file).T
            self.data_signal = Signal(data[0], data[1], self.settings)
        except OSError:
            raise Exception("{file} does not exist".format(file=file))

    def load_reference(self):
        """
        Load reference h(t)
        """
        file = self.settings.reference_file()
        self.settings.log("loading reference h(t)")
        try:
            data = loadtxt(file).T
            self.reference_signal = Signal(
                data[0], data[1], self.settings
            )
        except OSError:
            raise Exception("{file} does not exist".format(file=file))

    def load_coupling(self):
        """
        Load modelisation coupling data
        """
        file = self.settings.modelisation_file()
        self.settings.log("loading matlab modelisation data")
        try:
            self.modelisation = loadmat(file)
            coupling_values = self.modelisation[
                self.settings.coupling_name()
            ]
            self.coupling = [
                Signal(
                    self.modelisation["freq"][0],
                    coupling,
                    self.settings,
                )
                for coupling in coupling_values
            ]
        except OSError:
            raise Exception("{file} does not exist".format(file=file))

    def process_movement(self):
        """
        Clean and compute relative movement between the bench and the
        miror
        """
        self.settings.log("computing relative movement")

        self.movement = (
            (
                self.bench_movement * self.settings.calib_bench
                - self.mirror_movement * self.settings.calib_mirror
            )
            .detrend("linear")
            .low_pass_filter(5 * self.settings.vibration_frequency)
        )

    def compute_light(self):
        """
        Compute psd of the computed projection with current bench
        excitation
        """
        results = zeros(
            (
                len(self.coupling),
                len(
                    interpolate_abciss(
                        (self.movement.psd(), self.coupling[0].abs())
                    )
                ),
            )
        )

        # frequencies depends of psd result, which we do not have yet
        frequencies = zeros(results.shape[1])

        for index in range(len(self.coupling)):
            phase = (
                (index + 1) * 4 * pi / (self.settings.wavelength * 1e6)
                # * 1e6 sinon la courbe n'est plus du tout la même
            )

            factor_n = (self.movement * phase).sin().psd().sqrt()
            coupling_n = self.coupling[0].abs()
            factor_d = (self.movement * phase).cos().psd().sqrt()
            coupling_d = self.coupling[1].abs()

            factor_n, coupling_n, factor_d, coupling_d = interpolate(
                (factor_n, coupling_n, factor_d, coupling_d)
            )

            # no need to redefine it each time but simpler here
            frequencies = factor_n.x

            """ # from theoric equation
            results[index] = (
                sqrt(self.settings.scattering_factor[index])
                * self.settings.power_in
                / self.settings.power_out
                * (coupling_n * factor_n + coupling_d * factor_d).y
            )
            """
            results[index] = ( # from matlab
                sqrt(self.settings.scattering_factor[index])
                * self.settings.wavelength
                / (4 * pi)
                * (coupling_n**2 * factor_n + coupling_d**2 * factor_d)
                .sqrt()
                .y
            ) ** 2

        self.projection = Signal(
            frequencies,
            sqrt(sum(results)),
            self.settings,
        )
        return self.projection
